@@ -21,6 +21,12 @@
 # IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 # CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+"""Keycloak OpenID module.
+
+The module contains mainly the implementation of KeycloakOpenID class, the main
+class to handle authentication and token manipulation.
+"""
+
 import json
 
 from jose import jwt
@@ -52,8 +58,7 @@ from .urls_patterns import (
 
 
 class KeycloakOpenID:
-    """
-    Keycloak OpenID client.
+    """Keycloak OpenID client.
 
     :param server_url: Keycloak server url
     :param client_id: client id
@@ -62,6 +67,7 @@ class KeycloakOpenID:
     :param verify: True if want check connection SSL
     :param custom_headers: dict of custom header to pass to each HTML request
     :param proxies: dict of proxies to sent the request by.
+    :param timeout: connection timeout in seconds
     """
 
     def __init__(
@@ -73,19 +79,22 @@ class KeycloakOpenID:
         verify=True,
         custom_headers=None,
         proxies=None,
+        timeout=60,
     ):
+        """Init method."""
         self.client_id = client_id
         self.client_secret_key = client_secret_key
         self.realm_name = realm_name
         headers = custom_headers if custom_headers is not None else dict()
         self.connection = ConnectionManager(
-            base_url=server_url, headers=headers, timeout=60, verify=verify, proxies=proxies
+            base_url=server_url, headers=headers, timeout=timeout, verify=verify, proxies=proxies
         )
 
         self.authorization = Authorization()
 
     @property
     def client_id(self):
+        """Get client id."""
         return self._client_id
 
     @client_id.setter
@@ -94,6 +103,7 @@ class KeycloakOpenID:
 
     @property
     def client_secret_key(self):
+        """Get the client secret key."""
         return self._client_secret_key
 
     @client_secret_key.setter
@@ -102,6 +112,7 @@ class KeycloakOpenID:
 
     @property
     def realm_name(self):
+        """Get the realm name."""
         return self._realm_name
 
     @realm_name.setter
@@ -110,6 +121,7 @@ class KeycloakOpenID:
 
     @property
     def connection(self):
+        """Get connection."""
         return self._connection
 
     @connection.setter
@@ -118,6 +130,7 @@ class KeycloakOpenID:
 
     @property
     def authorization(self):
+        """Get authorization."""
         return self._authorization
 
     @authorization.setter
@@ -125,8 +138,7 @@ class KeycloakOpenID:
         self._authorization = value
 
     def _add_secret_key(self, payload):
-        """
-        Add secret key if exist.
+        """Add secret key if exists.
 
         :param payload:
         :return:
@@ -137,7 +149,7 @@ class KeycloakOpenID:
         return payload
 
     def _build_name_role(self, role):
-        """
+        """Build name of a role.
 
         :param role:
         :return:
@@ -145,7 +157,7 @@ class KeycloakOpenID:
         return self.client_id + "/" + role
 
     def _token_info(self, token, method_token_info, **kwargs):
-        """
+        """Getter for the token data.
 
         :param token:
         :param method_token_info:
@@ -160,29 +172,36 @@ class KeycloakOpenID:
         return token_info
 
     def well_known(self):
-        """The most important endpoint to understand is the well-known configuration
+        """Get the well_known object.
+
+        The most important endpoint to understand is the well-known configuration
         endpoint. It lists endpoints and other configuration options relevant to
         the OpenID Connect implementation in Keycloak.
 
         :return It lists endpoints and other configuration options relevant.
         """
-
         params_path = {"realm-name": self.realm_name}
         data_raw = self.connection.raw_get(URL_WELL_KNOWN.format(**params_path))
-
         return raise_error_from_response(data_raw, KeycloakGetError)
 
-    def auth_url(self, redirect_uri):
-        """
+    def auth_url(self, redirect_uri, scope="email", state=""):
+        """Get authorization URL endpoint.
 
-        http://openid.net/specs/openid-connect-core-1_0.html#AuthorizationEndpoint
-
-        :return:
+        :param redirect_uri: Redirect url to receive oauth code
+        :type redirect_uri: str
+        :param scope: Scope of authorization request, split with the blank space
+        :type: scope: str
+        :param state: State will be returned to the redirect_uri
+        :type: str
+        :returns: Authorization URL Full Build
+        :rtype: str
         """
         params_path = {
             "authorization-endpoint": self.well_known()["authorization_endpoint"],
             "client-id": self.client_id,
             "redirect-uri": redirect_uri,
+            "scope": scope,
+            "state": state,
         }
         return URL_AUTH.format(**params_path)
 
@@ -196,7 +215,8 @@ class KeycloakOpenID:
         totp=None,
         **extra
     ):
-        """
+        """Retrieve user token.
+
         The token endpoint is used to obtain tokens. Tokens can either be obtained by
         exchanging an authorization code or by supplying credentials directly depending on
         what flow is used. The token endpoint is also used to obtain new access tokens
@@ -229,10 +249,11 @@ class KeycloakOpenID:
 
         payload = self._add_secret_key(payload)
         data_raw = self.connection.raw_post(URL_TOKEN.format(**params_path), data=payload)
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return raise_error_from_response(data_raw, KeycloakPostError)
 
     def refresh_token(self, refresh_token, grant_type=["refresh_token"]):
-        """
+        """Refresh the user token.
+
         The token endpoint is used to obtain tokens. Tokens can either be obtained by
         exchanging an authorization code or by supplying credentials directly depending on
         what flow is used. The token endpoint is also used to obtain new access tokens
@@ -252,10 +273,11 @@ class KeycloakOpenID:
         }
         payload = self._add_secret_key(payload)
         data_raw = self.connection.raw_post(URL_TOKEN.format(**params_path), data=payload)
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return raise_error_from_response(data_raw, KeycloakPostError)
 
     def exchange_token(self, token: str, client_id: str, audience: str, subject: str) -> dict:
-        """
+        """Exchange user token.
+
         Use a token to obtain an entirely different token. See
         https://www.keycloak.org/docs/latest/securing_apps/index.html#_token-exchange
 
@@ -276,10 +298,11 @@ class KeycloakOpenID:
         }
         payload = self._add_secret_key(payload)
         data_raw = self.connection.raw_post(URL_TOKEN.format(**params_path), data=payload)
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return raise_error_from_response(data_raw, KeycloakPostError)
 
     def userinfo(self, token):
-        """
+        """Get the user info object.
+
         The userinfo endpoint returns standard claims about the authenticated user,
         and is protected by a bearer token.
 
@@ -288,30 +311,26 @@ class KeycloakOpenID:
         :param token:
         :return:
         """
-
         self.connection.add_param_headers("Authorization", "Bearer " + token)
         params_path = {"realm-name": self.realm_name}
-
         data_raw = self.connection.raw_get(URL_USERINFO.format(**params_path))
-
         return raise_error_from_response(data_raw, KeycloakGetError)
 
     def logout(self, refresh_token):
-        """
-        The logout endpoint logs out the authenticated user.
+        """Log out the authenticated user.
+
         :param refresh_token:
         :return:
         """
         params_path = {"realm-name": self.realm_name}
         payload = {"client_id": self.client_id, "refresh_token": refresh_token}
-
         payload = self._add_secret_key(payload)
         data_raw = self.connection.raw_post(URL_LOGOUT.format(**params_path), data=payload)
-
-        return raise_error_from_response(data_raw, KeycloakGetError, expected_codes=[204])
+        return raise_error_from_response(data_raw, KeycloakPostError, expected_codes=[204])
 
     def certs(self):
-        """
+        """Get certificates.
+
         The certificate endpoint returns the public keys enabled by the realm, encoded as a
         JSON Web Key (JWK). Depending on the realm settings there can be one or more keys enabled
         for verifying tokens.
@@ -325,7 +344,8 @@ class KeycloakOpenID:
         return raise_error_from_response(data_raw, KeycloakGetError)
 
     def public_key(self):
-        """
+        """Retrieve the public key.
+
         The public key is exposed by the realm page directly.
 
         :return:
@@ -335,7 +355,8 @@ class KeycloakOpenID:
         return raise_error_from_response(data_raw, KeycloakGetError)["public_key"]
 
     def entitlement(self, token, resource_server_id):
-        """
+        """Get entitlements from the token.
+
         Client applications can use a specific endpoint to obtain a special security token
         called a requesting party token (RPT). This token consists of all the entitlements
         (or permissions) for a user as a result of the evaluation of the permissions and
@@ -351,10 +372,11 @@ class KeycloakOpenID:
         if data_raw.status_code == 404:
             return raise_error_from_response(data_raw, KeycloakDeprecationError)
 
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return raise_error_from_response(data_raw, KeycloakGetError)  # pragma: no cover
 
     def introspect(self, token, rpt=None, token_type_hint=None):
-        """
+        """Introspect the user token.
+
         The introspection endpoint is used to retrieve the active state of a token.
         It is can only be invoked by confidential clients.
 
@@ -367,7 +389,6 @@ class KeycloakOpenID:
         :return:
         """
         params_path = {"realm-name": self.realm_name}
-
         payload = {"client_id": self.client_id, "token": token}
 
         if token_type_hint == "requesting_party_token":
@@ -380,11 +401,11 @@ class KeycloakOpenID:
         payload = self._add_secret_key(payload)
 
         data_raw = self.connection.raw_post(URL_INTROSPECT.format(**params_path), data=payload)
-
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return raise_error_from_response(data_raw, KeycloakPostError)
 
     def decode_token(self, token, key, algorithms=["RS256"], **kwargs):
-        """
+        """Decode user token.
+
         A JSON Web Key (JWK) is a JavaScript Object Notation (JSON) data
         structure that represents a cryptographic key.  This specification
         also defines a JWK Set JSON data structure that represents a set of
@@ -399,29 +420,25 @@ class KeycloakOpenID:
         :param algorithms:
         :return:
         """
-
         return jwt.decode(token, key, algorithms=algorithms, audience=self.client_id, **kwargs)
 
     def load_authorization_config(self, path):
-        """
-        Load Keycloak settings (authorization)
+        """Load Keycloak settings (authorization).
 
         :param path: settings file (json)
         :return:
         """
-        authorization_file = open(path, "r")
-        authorization_json = json.loads(authorization_file.read())
+        with open(path, "r") as fp:
+            authorization_json = json.load(fp)
+
         self.authorization.load_config(authorization_json)
-        authorization_file.close()
 
     def get_policies(self, token, method_token_info="introspect", **kwargs):
-        """
-        Get policies by user token
+        """Get policies by user token.
 
         :param token: user token
         :return: policies list
         """
-
         if not self.authorization.policies:
             raise KeycloakAuthorizationConfigError(
                 "Keycloak settings not found. Load Authorization Keycloak settings."
@@ -447,15 +464,13 @@ class KeycloakOpenID:
         return list(set(policies))
 
     def get_permissions(self, token, method_token_info="introspect", **kwargs):
-        """
-        Get permission by user token
+        """Get permission by user token.
 
         :param token: user token
         :param method_token_info: Decode token method
         :param kwargs: parameters for decode
         :return: permissions list
         """
-
         if not self.authorization.policies:
             raise KeycloakAuthorizationConfigError(
                 "Keycloak settings not found. Load Authorization Keycloak settings."
@@ -481,8 +496,7 @@ class KeycloakOpenID:
         return list(set(permissions))
 
     def uma_permissions(self, token, permissions=""):
-        """
-        Get UMA permissions by user token with requested permissions
+        """Get UMA permissions by user token with requested permissions.
 
         The token endpoint is used to retrieve UMA permissions from Keycloak. It can only be
         invoked by confidential clients.
@@ -493,7 +507,6 @@ class KeycloakOpenID:
         :param permissions: list of uma permissions list(resource:scope) requested by the user
         :return: permissions list
         """
-
         permission = build_permission_param(permissions)
 
         params_path = {"realm-name": self.realm_name}
@@ -506,12 +519,10 @@ class KeycloakOpenID:
 
         self.connection.add_param_headers("Authorization", "Bearer " + token)
         data_raw = self.connection.raw_post(URL_TOKEN.format(**params_path), data=payload)
-
         return raise_error_from_response(data_raw, KeycloakPostError)
 
     def has_uma_access(self, token, permissions):
-        """
-        Determine whether user has uma permissions with specified user token
+        """Determine whether user has uma permissions with specified user token.
 
         :param token: user token
         :param permissions: list of uma permissions (resource:scope)
@@ -521,7 +532,7 @@ class KeycloakOpenID:
         try:
             granted = self.uma_permissions(token, permissions)
         except (KeycloakPostError, KeycloakAuthenticationError) as e:
-            if e.response_code == 403:
+            if e.response_code == 403:  # pragma: no cover
                 return AuthStatus(
                     is_logged_in=True, is_authorized=False, missing_permissions=needed
                 )
@@ -537,7 +548,7 @@ class KeycloakOpenID:
             if not scopes:
                 needed.discard(resource)
                 continue
-            for scope in scopes:
+            for scope in scopes:  # pragma: no cover
                 needed.discard("{}#{}".format(resource, scope))
 
         return AuthStatus(
